/*
  ROOT-based implementation of reinterpretation top and W objects in SUS-16-049

  Returns number of reinterpreted merged Top, merged W, resolved Top

  Inputs/Outputs
  --------------------
    inMET   -> the event's MET
    inJets  -> the event's ak4 jets (primary jet collection; N_j in the analysis)
    inAk8   -> the event's ak8 jets
    inGenB  -> the event's generator-level b quarks
    rand    -> user-controlled TRandom3 instance, for reproducibility
    outNTop, outNW, outNRes are the output numbers of merged Top, merged W, and resolved Top, respectively

  User-defined object classes
  --------------------
    P4     -> a generic class for ak4 jets or gen particles. it can be a simple four-vector class having pt(), eta(), phi(), and mass() members
    AK8    -> a class for ak8 jets which must have in addition to a four-vector:
      soft drop variables sdmass() and minsubjetpt()
      an enum to indicate generator-level matching status, genCategory(), including GENTOP and GENW
      a pointer to a P4 of the matching generator-level particle, matchedGenParticle(--), passing GENTOP or GENW to request either
    ResTop -> a class for resolved Tops. must have:
      b, j2, j3 -> P4 for the three constituent ak4 jets
      wcand     -> P4 for the two-jet W system (= j2 + j3)
      topcand   -> P4 for the three-jet system (= b+j2+j3)
      bool overlaps(P4 jet) -> a function to check if b, j2, or j3 are the input jet (used for cross-cleaning)
      int nMatchedSubjets() -> a function returning the max number of the b/j2/j3 constituents which match to the decay products (b, W1, W2) of
        any single hadronically decaying generator-level top quark in the event, using deltaR < 0.4

  User-defined functions
  --------------------
    deltaR(thing1,thing2) returns the positive deltaR distance between the user-defined objects, deltaR = sqrt( deltaPhi(obj1,obj2)^2 + deltaEta(obj1,obj2)^2 )
    getTopWEff(EFFTYPE effType, float pt) should return the efficiency values from the provided pT-binned histograms
      the enum EFFTYPE can be TOPTAG, TOPMISTAG, WTAG, WMISTAG, RESTAG, RESMISTAG
    removeOverlaps(vector<thing1*>, vector<thing2*>, float DR) returns a 'cleaned' version of vector<thing1*> for which deltaR(any thing1, any thing2) > DR
    findNearest(thing1, vector<thing2*>, float DR) returns the nearest thing2 (an index into vector<thing2*>) to thing1 within deltaR(thing1, thing2) < DR
*/

  void fillEventSSRInfo(P4 * inMET, std::vector<P4*> inJets, std::vector<ak8*> inAk8, std::vector<P4*> inGenB, TRandom3 * rand, int &outNTop, int &outNW, int &outNRes) {

    outNTop = 0, outNW = 0, outNRes = 0;

    // merged top/W //
    std::vector<ak8*> candTopMatched, candWMatched, candTopUnmatched, candWUnmatched;
    std::vector<ak8*> passTopMatched, passWMatched, passTopUnmatched, passWUnmatched;
    for(auto * fj : inAk8){
      if(fj->pt() < 200 || TMath::Abs(fj->eta()) > 2.4 || fj->minsubjetpt() < 20 || fj->sdmass() < 50) continue;
      float fjpt = fj->pt(), fjeta = fj->eta();
      float sdmass = fj->softDropMass();

      bool isMatchedTop = fj->genCategory() & GenCategory::GENTOP;
      bool isMatchedW   = fj->genCategory() & GenCategory::GENW;

      // for W, in addition to the deltaR requirement for GenCategory, we defined the eff histograms to exclude nearby b quarks
      float mindrbW = 99.;
      for(auto * genb : inGenB){
        float tmpdr = deltaR(*genb, *fj);
        if(mintmpdr < drbW) { mindrbW = tmpdr; }
      }
      if(mindrbW < 0.8) { isMatchedW = false; }

      float toppt   = isMatchedTop ? fj->matchedGenParticle(GenCategory::GENTOP)->pt()  : fjpt; // gen pt for eff, reco pt for mistag
      float wpt     = isMatchedW   ? fj->matchedGenParticle(GenCategory::GENW)->pt()    : fjpt;

      float matchedEffTop   = getTopWEff(TOPTAG,    toppt);
      float matchedEffW     = getTopWEff(WTAG,      wpt);
      float unmatchedEffTop = getTopWEff(TOPMISTAG, toppt);
      float unmatchedEffW   = getTopWEff(WMISTAG,   wpt);

      // random tagging
      if(fjpt > 400 && sdmass > 110){
        if(isMatchedTop){
          candTopMatched.push_back(fj);
          if(rand->Uniform(0,1) < matchedEffTop) { passTopMatched.push_back(fj); }
        }else{
          candTopUnmatched.push_back(fj);
          if(rand->Uniform(0,1) < unmatchedEffTop) { passTopUnmatched.push_back(fj); }
        }
      }else if(fjpt > 200 && sdmass > 50 && sdmass < 110){
        if(isMatchedW){
          candWMatched.push_back(fj);
          if(rand->Uniform(0,1) < matchedEffW) { passWMatched.push_back(fj); }
        }else{
          candWUnmatched.push_back(fj);
          if(rand->Uniform(0,1) < unmatchedEffW) { passWUnmatched.push_back(fj); }
        }
      }

    }//ak8

    outNTop = passTopMatched.size()+passTopUnmatched.size();
    outNW   = passWMatched.size()+passWUnmatched.size();


    // resolved tops //
    std::vector<ak8*> passmerged;
    for(auto * fj : passTopMatched)   { passmerged.push_back(fj); }
    for(auto * fj : passTopUnmatched) { passmerged.push_back(fj); }
    for(auto * fj : passWMatched)     { passmerged.push_back(fj); }
    for(auto * fj : passWUnmatched)   { passmerged.push_back(fj); }

    // define b constituents
    std::vector<P4*> constituentb;
    std::vector<P4*> cleanedAK4 = removeOverlaps(inJets, passmerged, 0.8);
    for(auto * j : cleanedAK4){
      double nearDR;
      int near = findNearest(*j, inGenB, 0.4);
      if(near>=0) constituentb.push_back(j);
    }

    // form three-jet candidates
    std::vector<ResTop> allCands, cleanedCands;
    if(inJets.size() > 2) {
      for(auto * b : constituentb){
        for(unsigned i2 = 0 ; i2 < cleanedAK4.size()-1 ; ++i2){
          if(cleanedAK4[i2] == b) continue;
          for(unsigned i3 = i2+1 ; i3 < cleanedAK4.size(); ++i3){
            if(cleanedAK4[i3] == b) continue;
            ResTop tmpCand(b, cleanedAK4.at(i2), cleanedAK4.at(i3));
            if( TMath::Abs(tmpCand.wcand.mass() - 80.) > 40. || TMath::Abs(tmpCand.topcand.mass() - 173.) > 80.) continue;
            allCands.push_back(tmpCand);
          }
        }
      }
      // cross-clean overlaps using chi2
      auto reschi2 = [](const TopCand &a)->float {
        float topmass = a.topcand.mass();
        float wmass   = a.wcand.mass();
        float topDR   = deltaR(*(a.b), a.wcand);
        float topptDR = a.topcand.pt()*topDR;
        float WDR     = deltaR(*(a.j2), *(a.j3));
        float WptDR   = a.wcand.pt()*WDR;
        float sd0     = a.j3->pt()/(a.j2->pt()+a.j3->pt());
        float sd2     = sd0/std::pow(WDR, -2);
        float chi2 = (pow( (topmass-173.)/2, 2) + pow( (wmass-80.)/20, 2) + pow( (topptDR-250.)/50, 2) + pow( (WptDR-150.)/50, 2) + pow( (sd2-0.)/1., 2));
        return chi2;
      };
      std::sort(allCands.begin(),allCands.end(),[&](const TopCand &a, const TopCand &b){ return reschi2(a) < reschi2(b); }); // low to high chi2
      for(const auto &c : allCands){
        bool isOverlap = false;
        for(const auto &cleaned : cleanedCands){
          if(c.overlaps(cleaned)){
            isOverlap = true; break;
          }
        }
        if(!isOverlap) cleanedCands.push_back(c);
      }
    }//jets>2

    // random tagging
    std::vector<ResTop*> candResMatched, candResUnmatched;
    std::vector<ResTop*> passResMatched, passResUnmatched;
    for(auto &cand : cleanedCands){
      ResTop * resptr = &cand;
      if(cand.topcand.pt() < 100) continue;
      float matchedEffRes   = getTopWEff(RESTOPTAG,    cand.topcand.pt());
      float unmatchedEffRes = getTopWEff(RESTOPMISTAG, cand.topcand.pt());
      int nmatchsj = cand.nMatchedSubjets();
      if(nmatchsj >= 2){
        candResMatched.push_back(resptr);
        if(rand->Uniform(0,1) < matchedEffRes) passResMatched.push_back(resptr);
      }else{
        candResUnmatched.push_back(resptr);
        if(rand->Uniform(0,1) < unmatchedEffRes) passResUnmatched.push_back(resptr);
      }
    }

    outNRes = passResMatched.size()+passResUnmatched.size();

  }

